# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import fixtures
from testtools.matchers import Contains, Equals

from snapcraft import extractors
from snapcraft.internal import errors
from snapcraft.internal.pluginhandler import extract_metadata
from tests import fixture_setup, unit


class MetadataExtractionTestCase(unit.TestCase):
    def test_handled_file(self):
        open("test-metadata-file", "w").close()

        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(
                summary="test summary", description="test description"
            )

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        metadata = extract_metadata("test-part", "test-metadata-file", workdir=".")
        self.assertThat(metadata.get_summary(), Equals("test summary"))
        self.assertThat(metadata.get_description(), Equals("test description"))

    def test_unhandled_file(self):
        open("unhandled-file", "w").close()

        def _fake_extractor(file_path, workdir):
            raise extractors.UnhandledFileError(file_path, "fake")

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        raised = self.assertRaises(
            errors.UnhandledMetadataFileTypeError,
            extract_metadata,
            "test-part",
            "unhandled-file",
            workdir=".",
        )

        self.assertThat(raised.path, Equals("unhandled-file"))

    def test_extractor_missing_extract_function(self):
        fake_logger = fixtures.FakeLogger(level=logging.WARN)
        self.useFixture(fake_logger)

        open("unhandled-file", "w").close()

        def _fake_extractor(file_path, workdir):
            raise extractors.UnhandledFileError(file_path, "fake")

        self.useFixture(
            fixture_setup.FakeMetadataExtractor("fake", _fake_extractor, "not_extract")
        )

        raised = self.assertRaises(
            errors.UnhandledMetadataFileTypeError,
            extract_metadata,
            "test-part",
            "unhandled-file",
            workdir=".",
        )

        self.assertThat(raised.path, Equals("unhandled-file"))
        self.assertThat(
            fake_logger.output,
            Contains("'fake' doesn't include the 'extract' function"),
        )

    def test_extractor_returning_invalid_things(self):
        open("unhandled-file", "w").close()

        def _fake_extractor(file_path, workdir):
            return "I'm not metadata!"

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        raised = self.assertRaises(
            errors.InvalidExtractorValueError,
            extract_metadata,
            "test-part",
            "unhandled-file",
            workdir=".",
        )

        self.assertThat(raised.path, Equals("unhandled-file"))
        self.assertThat(raised.extractor_name, Equals("fake"))

    def test_missing_file(self):
        raised = self.assertRaises(
            errors.MissingMetadataFileError,
            extract_metadata,
            "test-part",
            "non-existent",
            workdir=".",
        )

        self.assertThat(raised.path, Equals("non-existent"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This plugin just dumps the content from a specified source.

This plugin uses the common plugin keywords as well as those for 'sources'.
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

In the cases where dumping the content needs some mangling or organizing
one would take advantage of the core functionalities available to plugins
such as: `filesets`, `stage`, `snap` and `organize`.
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class DumpPlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {},
            "required": ["source"],
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return set()

    def get_build_environment(self) -> Dict[str, str]:
        return dict()

    def get_build_commands(self) -> List[str]:
        return ['cp --archive --link --no-dereference . "${SNAPCRAFT_PART_INSTALL}"']



# -*- mode:python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

import fixtures
from testtools import TestCase
from testtools import matchers as m

from snapcraft.internal import deltas
from tests import fixture_setup


class BaseDeltaGenerationTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(fixture_setup.FakeTerminal())
        self.fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(self.fake_logger)

        self.delta_tool_path = "/usr/bin/xdelta3"
        self.workdir = self.useFixture(fixtures.TempDir()).path
        self.source_file = os.path.join(self.workdir, "source.snap")
        self.target_file = os.path.join(self.workdir, "target.snap")

        with open(self.source_file, "wb") as f:
            f.write(b"This is the source file.")
        with open(self.target_file, "wb") as f:
            f.write(b"This is the target file.")

        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.file_utils.get_snap_tool_path",
                side_effect=lambda x: os.path.join("/usr", "bin", x),
            )
        )

    def test_find_unique_file_name(self):
        tmp_delta = deltas.BaseDeltasGenerator(
            source_path=self.source_file,
            target_path=self.target_file,
            delta_format="xdelta3",
            delta_tool=self.delta_tool_path,
        )

        unique_file_name = tmp_delta.find_unique_file_name(tmp_delta.source_path)
        self.assertThat(unique_file_name, m.Equals(tmp_delta.source_path + "-0"))
        with open(unique_file_name, "wb") as f:
            f.write(b"tmp file.")

        self.assertThat(
            tmp_delta.find_unique_file_name(tmp_delta.source_path),
            m.Equals(tmp_delta.source_path + "-1"),
        )

    def test_not_set_delta_property_correctly(self):
        self.assertThat(
            lambda: deltas.BaseDeltasGenerator(
                source_path=self.source_file,
                target_path=self.target_file,
                delta_format="not-defined",
                delta_tool="/usr/bin/xdelta3",
            ),
            m.raises(deltas.errors.DeltaFormatOptionError),
        )
        exception = self.assertRaises(
            deltas.errors.DeltaFormatOptionError,
            deltas.BaseDeltasGenerator,
            source_path=self.source_file,
            target_path=self.target_file,
            delta_format="invalid-delta-format",
            delta_tool=self.delta_tool_path,
        )
        expected = """delta_format must be a option in ['xdelta3'].
for now delta_format='invalid-delta-format'"""
        self.assertThat(str(exception), m.Equals(expected))

    def test_file_existence_failed(self):
        class Tmpdelta(deltas.BaseDeltasGenerator):
            pass

        self.assertThat(
            lambda: deltas.BaseDeltasGenerator(
                source_path="invalid-source-file",
                target_path=self.target_file,
                delta_format="xdelta3",
                delta_tool="delta-gen-tool-path",
            ),
            m.raises(ValueError),
        )
        self.assertThat(
            lambda: deltas.BaseDeltasGenerator(
                source_path=self.source_file,
                target_path="invalid-target_file",
                delta_format="xdelta3",
                delta_tool="delta-gen-tool-path",
            ),
            m.raises(ValueError),
        )

    def test_subclass_not_implement_get_delta_cmd(self):
        tmp_delta = deltas.BaseDeltasGenerator(
            source_path=self.source_file,
            target_path=self.target_file,
            delta_format="xdelta3",
            delta_tool=self.delta_tool_path,
        )

        self.assertThat(
            lambda: tmp_delta.make_delta(is_for_test=True),
            m.raises(NotImplementedError),
        )

    def test_large_delta_raises_error(self):
        delta_file = os.path.join(self.workdir, "target.snap.delta")
        target_file = os.path.join(self.workdir, "target.snap")
        source_file = os.path.join(self.workdir, "source.snap")

        with open(delta_file, "wb") as delta, open(source_file, "wb") as source:
            delta.seek(999 - 1)
            delta.write(b"\0")
            source.seek(1000 - 1)
            source.write(b"\0")

        generator = deltas.BaseDeltasGenerator(
            source_path=source_file,
            target_path=target_file,
            delta_format="xdelta3",
            delta_tool=self.delta_tool_path,
        )

        self.assertThat(
            lambda: generator._check_delta_size_constraint(delta_file),
            m.raises(deltas.errors.DeltaGenerationTooBigError),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016, 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# LP: #1733579
from wsgiref import simple_server  # type: ignore

from pyramid import config


class BaseFakeServer:
    def __init__(self, server_address):
        super().__init__()
        self.server_address = server_address
        configurator = config.Configurator()
        self.configure(configurator)
        app = configurator.make_wsgi_app()
        self.server = simple_server.make_server(
            self.server_address[0], self.server_address[1], app
        )
        self.server_port = self.server.server_port
        self.socket = self.server.socket

    def configure(self, configurator):
        # To be defined in subclasses.
        raise NotImplementedError()

    def serve_forever(self):
        self.server.serve_forever()

    def shutdown(self):
        self.server.shutdown()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import textwrap
from unittest import mock

import fixtures
from testtools.matchers import (
    Contains,
    DirExists,
    Equals,
    FileContains,
    FileExists,
    Not,
)

import snapcraft
from snapcraft.internal import errors, lifecycle, pluginhandler, project_loader, steps
from snapcraft.internal.lifecycle._runner import _replace_in_part
from snapcraft.project import Project
from tests import fixture_setup

from . import LifecycleTestBase


class ExecutionTestCase(LifecycleTestBase):
    def test_replace_in_parts(self):
        class Options:
            def __init__(self):
                self.source = "$SNAPCRAFT_PART_INSTALL"

        class Plugin:
            def __init__(self):
                self.options = Options()

        class Part:
            def __init__(self):
                self.plugin = Plugin()
                self.part_source_dir = "/tmp"
                self.part_source_work_dir = "/tmp"
                self.part_build_dir = "/tmp"
                self.part_build_work_dir = "/tmp"
                self.part_install_dir = "/tmp"

        part = Part()
        new_part = _replace_in_part(part)

        self.assertThat(new_part.plugin.options.source, Equals(part.part_install_dir))

    @mock.patch("snapcraft.repo.snaps.install_snaps")
    def test_dependency_is_staged_when_required(self, mock_install_build_snaps):
        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  part1:
                    plugin: nil
                  part2:
                    plugin: nil
                    after:
                      - part1
                """
            )
        )

        lifecycle.execute(steps.PULL, project_config, part_names=["part2"])

        self.assertThat(
            self.fake_logger.output,
            Contains("'part2' has dependencies that need to be staged: part1"),
        )

    @mock.patch("snapcraft.repo.snaps.install_snaps")
    def test_no_exception_when_dependency_is_required_but_already_staged(
        self, mock_install_build_snaps
    ):
        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  part1:
                    plugin: nil
                  part2:
                    plugin: nil
                    after:
                      - part1
                """
            )
        )

        def _fake_should_step_run(self, step, force=False):
            return self.name != "part1"

        with mock.patch.object(
            pluginhandler.PluginHandler, "should_step_run", _fake_should_step_run
        ):
            lifecycle.execute(steps.PULL, project_config, part_names=["part2"])

        self.assertThat(self.fake_logger.output, Contains("Pulling part2"))
        self.assertThat(self.fake_logger.output, Not(Contains("Pulling part1")))

    def test_dirty_stage_part_with_built_dependent_raises(self):
        # Set the option to error on dirty/outdated steps
        with snapcraft.config.CLIConfig() as cli_config:
            cli_config.set_outdated_step_action(
                snapcraft.config.OutdatedStepAction.ERROR
            )

        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  part1:
                    plugin: nil
                  part2:
                    plugin: nil
                    after: [part1]
                """
            )
        )

        # Stage dependency
        lifecycle.execute(steps.STAGE, project_config, part_names=["part1"])
        # Build dependent
        lifecycle.execute(steps.BUILD, project_config, part_names=["part2"])

        def _fake_dirty_report(self, step):
            if step == steps.STAGE:
                return pluginhandler.DirtyReport(
                    dirty_properties={"foo"}, dirty_project_options={"bar"}
                )
            return None

        # Should stage no problem
        with mock.patch.object(
            pluginhandler.PluginHandler, "get_dirty_report", _fake_dirty_report
        ):
            lifecycle.execute(steps.STAGE, project_config, part_names=["part1"])

        # Reset logging since we only care about the following
        self.fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(self.fake_logger)

        # Should raise an error since part2 is now dirty
        raised = self.assertRaises(
            errors.StepOutdatedError, lifecycle.execute, steps.BUILD, project_config
        )

        output = self.fake_logger.output.split("\n")
        part1_output = [line.strip() for line in output if "part1" in line]
        self.assertThat(part1_output, Equals(["Skipping pull part1 (already ran)"]))

        self.assertThat(raised.step, Equals(steps.PULL))
        self.assertThat(raised.part, Equals("part2"))
        self.assertThat(raised.report, Equals("A dependency has changed: 'part1'\n"))

    @mock.patch("snapcraft.repo.snaps.install_snaps")
    def test_dirty_build_raises(self, mock_install_build_snaps):
        # Set the option to error on dirty/outdated steps
        with snapcraft.config.CLIConfig() as cli_config:
            cli_config.set_outdated_step_action(
                snapcraft.config.OutdatedStepAction.ERROR
            )

        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  part1:
                    plugin: nil
                """
            )
        )

        # Build it.
        lifecycle.execute(steps.BUILD, project_config)

        # Reset logging since we only care about the following
        self.fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(self.fake_logger)

        def _fake_dirty_report(self, step):
            if step == steps.BUILD:
                return pluginhandler.DirtyReport(dirty_properties={"foo", "bar"})
            return None

        # Should catch that the part needs to be rebuilt and raise an error.
        with mock.patch.object(
            pluginhandler.PluginHandler, "get_dirty_report", _fake_dirty_report
        ):
            raised = self.assertRaises(
                errors.StepOutdatedError, lifecycle.execute, steps.BUILD, project_config
            )

        self.assertThat(
            self.fake_logger.output, Equals("Skipping pull part1 (already ran)\n")
        )

        self.assertThat(raised.step, Equals(steps.BUILD))
        self.assertThat(raised.part, Equals("part1"))
        self.assertThat(
            raised.report,
            Equals("The 'bar' and 'foo' part properties appear to have changed.\n"),
        )
        self.assertThat(raised.parts_names, Equals("part1"))

    @mock.patch("snapcraft.repo.snaps.install_snaps")
    def test_dirty_pull_raises(self, mock_install_build_snaps):
        # Set the option to error on dirty/outdated steps
        with snapcraft.config.CLIConfig() as cli_config:
            cli_config.set_outdated_step_action(
                snapcraft.config.OutdatedStepAction.ERROR
            )

        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  part1:
                    plugin: nil
                """
            )
        )

        # Pull it.
        lifecycle.execute(steps.PULL, project_config)

        # Reset logging since we only care about the following
        self.fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(self.fake_logger)

        def _fake_dirty_report(self, step):
            if step == steps.PULL:
                return pluginhandler.DirtyReport(dirty_project_options={"foo", "bar"})
            return None

        # Should catch that the part needs to be re-pulled and raise an error.
        with mock.patch.object(
            pluginhandler.PluginHandler, "get_dirty_report", _fake_dirty_report
        ):
            raised = self.assertRaises(
                errors.StepOutdatedError, lifecycle.execute, steps.PULL, project_config
            )

        self.assertThat(self.fake_logger.output, Equals(""))

        self.assertThat(raised.step, Equals(steps.PULL))
        self.assertThat(raised.part, Equals("part1"))
        self.assertThat(
            raised.report,
            Equals("The 'bar' and 'foo' project options appear to have changed.\n"),
        )

    @mock.patch.object(snapcraft.BasePlugin, "enable_cross_compilation")
    @mock.patch("snapcraft.repo.Repo.install_build_packages")
    @mock.patch("snapcraft.repo.snaps.install_snaps")
    def test_pull_is_dirty_if_target_arch_changes(
        self,
        mock_install_build_snaps,
        mock_install_build_packages,
        mock_enable_cross_compilation,
    ):
        # Set the option to error on dirty/outdated steps
        with snapcraft.config.CLIConfig() as cli_config:
            cli_config.set_outdated_step_action(
                snapcraft.config.OutdatedStepAction.ERROR
            )

        mock_install_build_packages.return_value = []
        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  part1:
                    plugin: nil
                """
            )
        )

        project = Project(
            snapcraft_yaml_file_path=self.snapcraft_yaml_file_path,
            target_deb_arch="amd64",
        )
        project_config = project_loader.load_config(project)
        # Pull it with amd64
        lifecycle.execute(steps.PULL, project_config)

        # Reset logging since we only care about the following
        self.fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(self.fake_logger)

        project = Project(
            snapcraft_yaml_file_path=self.snapcraft_yaml_file_path,
            target_deb_arch="armhf",
        )
        project_config = project_loader.load_config(project)
        # Pull it again with armhf. Should catch that the part needs to be
        # re-pulled due to the change in target architecture and raise an
        # error.
        raised = self.assertRaises(
            errors.StepOutdatedError, lifecycle.execute, steps.PULL, project_config
        )

        self.assertThat(
            self.fake_logger.output, Contains("Setting target machine to 'armhf'")
        )

        self.assertThat(raised.step, Equals(steps.PULL))
        self.assertThat(raised.part, Equals("part1"))
        self.assertThat(
            raised.report,
            Equals("The 'deb_arch' project option appears to have changed.\n"),
        )

    def test_prime_excludes_internal_snapcraft_dir(self):
        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  test-part:
                    plugin: nil
                """
            )
        )
        lifecycle.execute(steps.PRIME, project_config)
        self.assertThat(
            os.path.join(steps.PRIME.name, "snap", ".snapcraft"), Not(DirExists())
        )

    def test_non_prime_and_no_version(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path, version=None)
        snapcraft_yaml.data["adopt-info"] = "test-part"
        snapcraft_yaml.update_part(
            "test-part",
            {"plugin": "nil", "override-build": "snapcraftctl set-version 1.0"},
        )
        self.useFixture(snapcraft_yaml)

        project = Project(
            snapcraft_yaml_file_path=snapcraft_yaml.snapcraft_yaml_file_path
        )
        project_config = project_loader.load_config(project)

        # This should not fail
        lifecycle.execute(steps.PULL, project_config)


class CleanTestCase(LifecycleTestBase):
    def test_clean_removes_global_state(self):
        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  test-part:
                    plugin: nil
                """
            )
        )
        lifecycle.execute(steps.PULL, project_config)
        lifecycle.clean(project_config.project, parts=None)
        self.assertThat(os.path.join("snap", ".snapcraft"), Not(DirExists()))

    @mock.patch("snapcraft.internal.mountinfo.MountInfo.for_root")
    def test_clean_leaves_prime_alone_for_tried(self, mock_for_root):
        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  test-part:
                    plugin: nil
                """
            )
        )
        lifecycle.execute(steps.PRIME, project_config)
        lifecycle.clean(project_config.project, parts=None)
        self.assertThat(
            steps.PRIME.name,
            DirExists(),
            "Expected prime directory to remain after cleaning for tried snap",
        )


class RecordSnapcraftYamlTestCase(LifecycleTestBase):
    def test_prime_without_build_info_does_not_record(self):
        self.useFixture(fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_INFO", None))
        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  test-part:
                    plugin: nil
                """
            )
        )
        lifecycle.execute(steps.PRIME, project_config)
        for file_name in ("snapcraft.yaml", "manifest.yaml"):
            self.assertThat(
                os.path.join(steps.PRIME.name, "snap", file_name), Not(FileExists())
            )

    def test_prime_with_build_info_records_snapcraft_yaml(self):
        self.useFixture(fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_INFO", "1"))
        project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  test-part:
                    plugin: nil
                """
            ),
            snap_type="type: app",
        )
        lifecycle.execute(steps.PRIME, project_config)

        expected = textwrap.dedent(
            """\
            name: test
            base: core18
            version: "1.0"
            summary: test
            description: test
            confinement: strict
            grade: stable
            type: app

            parts:
              test-part:
                plugin: nil

            """
        )

        self.assertThat(
            os.path.join(steps.PRIME.name, "snap", "snapcraft.yaml"),
            FileContains(expected),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import glob
import os

import xdg
from testtools.matchers import Equals, HasLength

from tests import integration


class StagePackageCacheTestCase(integration.TestCase):
    def test_stage_package_gets_cached(self):
        self.run_snapcraft(["pull", "oneflatwithstagepackages"], "dump")

        # Verify the 'hello' deb package was cached.
        cache_dir = os.path.join(
            xdg.BaseDirectory.xdg_cache_home, "snapcraft", "stage-packages", "apt"
        )
        archive_dir = os.path.join("var", "cache", "apt", "archives")
        cached = glob.glob(os.path.join(cache_dir, "*", archive_dir, "hello*"))
        self.assertThat(cached, HasLength(1))
        cached_deb = cached[0]

        staged = glob.glob(
            os.path.join(
                "parts", "oneflatwithstagepackages", "ubuntu", "download", "hello*"
            )
        )
        self.assertThat(staged, HasLength(1))
        staged_deb = staged[0]

        # Verify that the staged and cached debs are the same file (hard
        # linked) by comparing inodes.
        cached_deb_inode = os.stat(cached_deb).st_ino
        self.assertThat(cached_deb_inode, Equals(os.stat(staged_deb).st_ino))

        # Now clean the part and pull again.
        self.run_snapcraft("clean", "dump")
        self.run_snapcraft(["pull", "oneflatwithstagepackages"], "dump")

        # Verify that the staged deb is _still_ the same one from the cache.
        self.assertThat(cached_deb_inode, Equals(os.stat(staged_deb).st_ino))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import json
import os
import subprocess
import sys
import tempfile
import textwrap
import time
from typing import Any, Callable, Dict

from snapcraft.internal import common, errors, steps


class Runner:
    """The Runner class is responsible for orchestrating scriptlets."""

    # FIXME: Need to quote builtin_functions typing because of
    # https://github.com/python/typing/issues/259 which is fixed in Python
    # 3.5.3.
    def __init__(
        self,
        *,
        part_properties: Dict[str, Any],
        partdir: str,
        sourcedir: str,
        builddir: str,
        stagedir: str,
        primedir: str,
        builtin_functions: Dict[str, Callable[..., None]],
        env_generator: Callable[..., str],
        shell: str = "/bin/sh",
        shell_flags: str = "set -ex",
    ) -> None:
        """Create a new Runner.
        :param dict part_properties: YAML properties set for this part.
        :param str partdir: the root directory for this part.
        :param str sourcedir: The source directory for this part.
        :param str builddir: The build directory for this part.
        :param str stagedir: The staging area.
        :param str primedir: The priming area.
        :param dict builtin_functions: Dict of builtin function names to
                                       actual callables.
        :param str env_generator: a callable to provide environment settings
                                  for the part.
        """
        self._partdir = partdir
        self._sourcedir = sourcedir
        self._builddir = builddir
        self._stagedir = stagedir
        self._primedir = primedir
        self._builtin_functions = builtin_functions
        self._env_generator = env_generator

        self._override_pull_scriptlet = part_properties.get("override-pull")
        self._override_build_scriptlet = part_properties.get("override-build")
        self._override_stage_scriptlet = part_properties.get("override-stage")
        self._override_prime_scriptlet = part_properties.get("override-prime")

        self._shell = shell
        self._shell_flags = shell_flags

    def pull(self) -> None:
        """Run override-pull scriptlet."""
        if self._override_pull_scriptlet:
            self._run_scriptlet(
                "override-pull",
                self._override_pull_scriptlet,
                self._sourcedir,
                steps.PULL,
            )

    def build(self) -> None:
        """Run override-build scriptlet."""
        if self._override_build_scriptlet:
            self._run_scriptlet(
                "override-build",
                self._override_build_scriptlet,
                self._builddir,
                steps.BUILD,
            )

    def stage(self) -> None:
        """Run override-stage scriptlet."""
        if self._override_stage_scriptlet:
            self._run_scriptlet(
                "override-stage",
                self._override_stage_scriptlet,
                self._stagedir,
                steps.STAGE,
            )

    def prime(self) -> None:
        """Run override-prime scriptlet."""
        if self._override_prime_scriptlet:
            self._run_scriptlet(
                "override-prime",
                self._override_prime_scriptlet,
                self._primedir,
                steps.PRIME,
            )

    def _run_scriptlet(
        self, scriptlet_name: str, scriptlet: str, workdir: str, step: steps.Step
    ) -> None:
        if common.is_snap():
            # Since the snap is classic, there is no $PATH pointing into the snap, which
            # means snapcraftctl won't be found. We can't use aliases since they don't
            # persist into subshells. However, we know that snapcraftctl lives in its own
            # directory, so adding that to the PATH should have no ill side effects.
            snapcraftctl_env = 'export PATH="$PATH:$SNAP/bin/scriptlet-bin"'
        else:
            snapcraftctl_env = ""

        with tempfile.TemporaryDirectory(dir=self._partdir) as tempdir:
            call_fifo = _NonBlockingRWFifo(os.path.join(tempdir, "function_call"))
            feedback_fifo = _NonBlockingRWFifo(os.path.join(tempdir, "call_feedback"))

            # snapcraftctl only works consistently if it's using the exact same
            # interpreter as that used by snapcraft itself, thus the definition
            # of SNAPCRAFT_INTERPRETER.
            script = textwrap.dedent(
                """\
                set -e
                export SNAPCRAFTCTL_CALL_FIFO={call_fifo}
                export SNAPCRAFTCTL_FEEDBACK_FIFO={feedback_fifo}
                export SNAPCRAFT_INTERPRETER={interpreter}
                {snapcraftctl_env}

                {env}

                {shell_flags}

                {scriptlet}"""
            ).format(
                shell_flags=self._shell_flags,
                interpreter=sys.executable,
                call_fifo=call_fifo.path,
                feedback_fifo=feedback_fifo.path,
                scriptlet=scriptlet,
                snapcraftctl_env=snapcraftctl_env,
                env=self._env_generator(step),
            )

            with tempfile.TemporaryFile(mode="w+") as script_file:
                print(script, file=script_file)
                script_file.flush()
                script_file.seek(0)

                process = subprocess.Popen(
                    [self._shell], stdin=script_file, cwd=workdir
                )

            status = None
            try:
                while status is None:
                    function_call = call_fifo.read()
                    if function_call:
                        # Handle the function and let caller know that function
                        # call has been handled (must contain at least a
                        # newline, anything beyond is considered an error by
                        # snapcraftctl)
                        self._handle_builtin_function(
                            scriptlet_name, function_call.strip()
                        )
                        feedback_fifo.write("\n")

                    status = process.poll()

                    # Don't loop TOO busily
                    time.sleep(0.1)
            except Exception as error:
                feedback_fifo.write(f"{error!s}\n")
                raise error
            finally:
                call_fifo.close()
                feedback_fifo.close()

            if process.returncode != 0:
                raise errors.ScriptletRunError(
                    scriptlet_name=scriptlet_name, code=status
                )

    def _handle_builtin_function(self, scriptlet_name, function_call) -> None:
        try:
            function_json = json.loads(function_call)
        except json.decoder.JSONDecodeError as e:
            # This means a snapcraft developer messed up adding a new
            # snapcraftctl function. Should never be encountered in real life.
            raise ValueError(
                "{!r} scriptlet called a function with invalid json: "
                "{}".format(scriptlet_name, function_call)
            ) from e

        try:
            function_name = function_json["function"]
            function_args = function_json["args"]
        except KeyError as e:
            # This means a snapcraft developer messed up adding a new
            # snapcraftctl function. Should never be encountered in real life.
            raise ValueError(
                "{!r} scriptlet missing expected json field {!s} in args for "
                "function call {!r}: {}".format(
                    scriptlet_name, e, function_name, function_args
                )
            ) from e

        try:
            function = self._builtin_functions[function_name]
        except KeyError as e:
            # This means a snapcraft developer messed up adding a new
            # snapcraftctl function. Should never be encountered in real life.
            raise ValueError(
                "{!r} scriptlet called an undefined builtin function: "
                "{}".format(scriptlet_name, function_name)
            ) from e

        function(**function_args)


class _NonBlockingRWFifo:
    def __init__(self, path) -> None:
        os.mkfifo(path)
        self.path = path

        # Using RDWR for every FIFO just so we can open them reliably whenever
        # (i.e. write-only FIFOs can't be opened successfully until the reader
        # is in place)
        self._fd = os.open(self.path, os.O_RDWR | os.O_NONBLOCK)

    def read(self) -> str:
        total_read = ""
        with contextlib.suppress(BlockingIOError):
            value = os.read(self._fd, 1024)
            while value:
                total_read += value.decode(sys.getfilesystemencoding())
                value = os.read(self._fd, 1024)
        return total_read

    def write(self, data: str) -> int:
        return os.write(self._fd, data.encode(sys.getfilesystemencoding()))

    def close(self) -> None:
        if self._fd is not None:
            os.close(self._fd)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from typing import Dict  # noqa: F401
from typing import FrozenSet, List

from snapcraft.internal import elf, errors
from snapcraft.project import Project

logger = logging.getLogger(__name__)


class PartPatcher:
    """Takes care of patching files in a part if necessary."""

    def __init__(
        self,
        *,
        elf_files: FrozenSet[elf.ElfFile],
        project: Project,
        snap_base_path: str,
        stage_packages: List[str],
    ) -> None:
        """Initialize PartPatcher.

        :param elf_files: the list of elf files to analyze.
        :param project: the project instance from the part.
        :param snap_base_path: the root path of the snap during runtime,
                               necessary when using a in-snap libc6 provided
                               as a stage-packages entry.
        :param stage_packages: the stage-packages a part is set to use.
        """
        self._elf_files = elf_files
        self._project = project
        self._snap_base_path = snap_base_path
        # If libc6 is staged, to avoid symbol mixups we will resort to
        # glibc mangling.
        self._is_libc6_staged = "libc6" in stage_packages

    def _get_glibc_compatibility(self, linker_version: str) -> Dict[str, str]:
        linker_incompat = dict()  # type: Dict[str, str]
        for elf_file in self._elf_files:
            if not elf_file.is_linker_compatible(linker_version=linker_version):
                linker_incompat[elf_file.path] = elf_file.get_required_glibc()
        return linker_incompat

    def _get_preferred_patchelf_path(self):
        # TODO revisit if we need to support variations and permutations
        #  of this
        staged_patchelf_path = os.path.join(self._project.stage_dir, "bin", "patchelf")
        if not os.path.exists(staged_patchelf_path):
            staged_patchelf_path = None
        return staged_patchelf_path

    def _patch(self, dynamic_linker: str) -> None:
        preferred_patchelf_path = self._get_preferred_patchelf_path()

        elf_patcher = elf.Patcher(
            dynamic_linker=dynamic_linker,
            root_path=self._project.prime_dir,
            preferred_patchelf_path=preferred_patchelf_path,
        )

        # Patching all files instead of a subset of them to ensure the
        # environment is consistent and the chain of dlopens that may
        # happen remains sane.
        for elf_file in self._elf_files:
            try:
                elf_patcher.patch(elf_file=elf_file)
            except errors.PatcherError as patch_error:
                logger.warning(
                    "An attempt to patch {!r} so that it would work "
                    "correctly in diverse environments was made and failed. "
                    "To disable this behavior set "
                    "`build-attributes: [no-patchelf]` for the part.".format(
                        elf_file.path
                    )
                )
                raise patch_error

    def _verify_compat(self) -> None:
        if self._project._snap_meta.base is None:
            return

        linker_version = self._project._get_linker_version_for_base(
            self._project._snap_meta.base
        )
        linker_incompat = self._get_glibc_compatibility(linker_version)
        logger.debug(
            "List of files incompatible with {!r}: {!r}".format(
                linker_version, linker_incompat
            )
        )
        # Even though we do this in patch, it does not hurt to check again
        if linker_incompat and not self._is_libc6_staged:
            raise errors.IncompatibleBaseError(
                base=self._project._snap_meta.base,
                linker_version=linker_version,
                file_list=linker_incompat,
            )

    def patch(self) -> None:
        """Executes the patching process for elf_files.

        First it verifies there are no GLIBC mismatches, if any are found,
        libc6 must be part of stage-packages. The verification is only
        performed if the core base is not compatible with the host the
        snapcraft build is taking place on (e.g.; building on 18.04 while
        targeting 'core' as the core base).

        If the snapcraft project is not a classic confined one and there
        are no GLIBC mismatches, then this method returns with no modifications
        to elf_files. However, if one of those cases are true, the
        required dynamic linker is retrieved and the necessary elf files
        in elf_files are patched to behave accordinginly in the environment by
        means of setting the correct interpreter and rpaths (not runpaths).

        :raises errors.SnapcraftMissingLinkerInBaseError:
            if the linker within the core base cannot be found.
        :raises errors.StagePackageMissingError:
            if there are libc6 mismatches and libc6 is not in stage-packages.
        :raises errors.SnapcraftEnvironementError:
            if something is horribly wrong.
        """
        # Just return if base is None or a static base and libc6 has not been staged.
        if self._project._snap_meta.base is None or (
            self._project.is_static_base(self._project._snap_meta.base)
            and not self._is_libc6_staged
        ):
            return
        if not (
            self._project.is_static_base(self._project._snap_meta.base)
            or self._project.is_host_compatible_with_base(self._project._snap_meta.base)
        ):
            logger.debug("Host is not compatible with base")
            self._verify_compat()
        logger.debug("Is libc6 in stage-packages: {!r}".format(self._is_libc6_staged))

        if self._is_libc6_staged:
            dynamic_linker = elf.find_linker(
                root_path=self._project.prime_dir, snap_base_path=self._snap_base_path
            )
            logger.warning(
                "libc6 has been staged into the snap: only do this if you know what "
                "what you are doing."
            )
        else:
            dynamic_linker = self._project.get_core_dynamic_linker(
                self._project._snap_meta.base, expand=False
            )

        logger.debug("Dynamic linker set to {!r}".format(dynamic_linker))
        self._patch(dynamic_linker)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import Equals, FileExists

from tests import integration


class DebSourceTestCase(integration.TestCase):
    def test_stage_deb(self):
        self.copy_project_to_cwd("deb-hello")
        self.run_snapcraft(["stage", "deb"])

        self.assertThat(os.path.join(self.stage_dir, "bin", "hello"), FileExists())
        self.assertThat(
            os.path.join(self.stage_dir, "usr", "bin", "world"), FileExists()
        )

    # Regression test for LP: #1634813
    def test_stage_deb_with_symlink(self):
        self.copy_project_to_cwd("deb-with-symlink")
        self.run_snapcraft(["stage", "deb-with-symlink"])

        target = os.path.join(self.stage_dir, "target")
        symlink = os.path.join(self.stage_dir, "symlink")
        self.assertThat(target, FileExists())
        self.assertThat(symlink, FileExists())
        self.assertTrue(os.path.islink(symlink))
        self.assertThat(os.readlink(symlink), Equals("target"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Set

from snapcraft.plugins.v2 import nil


class PluginImpl(nil.NilPlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["foo"] = {"type": "string"}
        return schema

    def get_build_packages(self) -> Set[str]:
        build_packages = super().get_build_packages()
        build_packages.add("gcc")
        return build_packages

    def get_build_commands(self) -> List[str]:
        commands = super().get_build_commands()
        commands.append("mkdir -p ${SNAPCRAFT_PART_INSTALL}/bin")
        commands.append("gcc hello.c -o ${SNAPCRAFT_PART_INSTALL}/bin/hello")
        return commands



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import distutils.util
import os
import sys
from typing import Optional

from snapcraft.internal.errors import XAttributeError, XAttributeTooLongError


def _get_snapcraft_xattr_key(snapcraft_key: str) -> str:
    return f"user.snapcraft.{snapcraft_key}"


def _read_snapcraft_xattr(path: str, snapcraft_key: str) -> Optional[str]:
    if sys.platform != "linux":
        raise RuntimeError("xattr support only available for Linux")

    # Extended attributes do not apply to symlinks.
    if os.path.islink(path):
        return None

    key = _get_snapcraft_xattr_key(snapcraft_key)
    try:
        value = os.getxattr(path, key)
    except OSError as error:
        # No label present with:
        # OSError: [Errno 61] No data available: b'<path>'
        if error.errno == 61:
            return None

        # Chain unknown variants of OSError.
        raise XAttributeError(action="read", key=key, path=path) from error

    return value.decode().strip()


def _write_snapcraft_xattr(path: str, snapcraft_key: str, value: str) -> None:
    if sys.platform != "linux":
        raise RuntimeError("xattr support only available for Linux")

    # Extended attributes do not apply to symlinks.
    if os.path.islink(path):
        return

    key = _get_snapcraft_xattr_key(snapcraft_key)

    try:
        os.setxattr(path, key, value.encode())
    except OSError as error:
        # Label is too long for filesystem:
        # OSError: [Errno 7] Argument list too long: b'<path>'
        if error.errno == 7:
            raise XAttributeTooLongError(path=path, key=key, value=value)

        # Chain unknown variants of OSError.
        raise XAttributeError(action="write", key=key, path=path) from error


def read_origin_stage_package(path: str) -> Optional[str]:
    """Read origin stage package."""
    try:
        return _read_snapcraft_xattr(path, "origin_stage_package")
    except XAttributeError:
        # Ignore error if origin stage package not required.
        if distutils.util.strtobool(os.environ.get("SNAPCRAFT_BUILD_INFO", "n")):
            raise
        return None


def write_origin_stage_package(path: str, value: str) -> None:
    """Write origin stage package."""
    try:
        _write_snapcraft_xattr(path, "origin_stage_package", value)
    except XAttributeError:
        # Ignore error if origin stage package not required.
        if distutils.util.strtobool(os.environ.get("SNAPCRAFT_BUILD_INFO", "n")):
            raise



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from ._plugin import PluginV1  # noqa: F401

# The plugin code requires imports that are platform specific.
if sys.platform == "linux":
    from .ant import AntPlugin  # noqa: F401
    from .autotools import AutotoolsPlugin  # noqa: F401
    from .catkin import CatkinPlugin  # noqa: F401
    from .catkin_tools import CatkinToolsPlugin  # noqa: F401
    from .cmake import CMakePlugin  # noqa: F401
    from .colcon import ColconPlugin  # noqa: F401
    from .conda import CondaPlugin  # noqa: F401
    from .crystal import CrystalPlugin  # noqa: F401
    from .dotnet import DotNetPlugin  # noqa: F401
    from .dump import DumpPlugin  # noqa: F401
    from .flutter import FlutterPlugin  # noqa: F401
    from .go import GoPlugin  # noqa: F401
    from .godeps import GodepsPlugin  # noqa: F401
    from .gradle import GradlePlugin  # noqa: F401
    from .kbuild import KBuildPlugin  # noqa: F401
    from .kernel import KernelPlugin  # noqa: F401
    from .make import MakePlugin  # noqa: F401
    from .maven import MavenPlugin  # noqa: F401
    from .meson import MesonPlugin  # noqa: F401
    from .nil import NilPlugin  # noqa: F401
    from .nodejs import NodePlugin  # noqa: F401
    from .plainbox_provider import PlainboxProviderPlugin  # noqa: F401
    from .python import PythonPlugin  # noqa: F401
    from .qmake import QmakePlugin  # noqa: F401
    from .ruby import RubyPlugin  # noqa: F401
    from .rust import RustPlugin  # noqa: F401
    from .scons import SconsPlugin  # noqa: F401
    from .waf import WafPlugin  # noqa: F401



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import snapcraft.yaml_utils.errors
from snapcraft.internal import errors
from tests import fixture_setup

from . import CommandBaseTestCase


class CleanbuildCase(CommandBaseTestCase):
    def test_cleanbuild_raises_when_yaml_is_valid(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path, name="cleanbuild")
        snapcraft_yaml.update_part("part1", dict(plugin="nil"))
        self.useFixture(snapcraft_yaml)

        self.assertRaises(
            errors.SnapcraftEnvironmentError, self.run_command, ["cleanbuild"]
        )

    def test_cleanbuild_raises_when_yaml_is_invalid(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path, name="cleanbuild")
        snapcraft_yaml.update_part("part1", dict(sources="boo"))
        self.useFixture(snapcraft_yaml)

        self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            self.run_command,
            ["cleanbuild"],
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.project._project_options import _find_machine


class TestFindMachine:

    scenarios = [
        ("x86_64", dict(machine="x86_64", expected_machine="x86_64")),
        ("amd64", dict(machine="amd64", expected_machine="x86_64")),
        ("i386", dict(machine="i386", expected_machine="i686")),
        ("i686", dict(machine="i686", expected_machine="i686")),
        ("armhf", dict(machine="armhf", expected_machine="armv7l")),
        ("arm", dict(machine="arm", expected_machine="armv7l")),
        ("aarch64", dict(machine="aarch64", expected_machine="aarch64")),
        ("arm64", dict(machine="arm64", expected_machine="aarch64")),
        ("ppc64el", dict(machine="ppc64el", expected_machine="ppc64le")),
        ("ppc", dict(machine="powerpc", expected_machine="ppc")),
        ("s390x", dict(machine="s390x", expected_machine="s390x")),
    ]

    def test_find_machine(self, machine, expected_machine):
        assert _find_machine(machine) == expected_machine



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import glob
import os

from testtools.matchers import DirExists, Not

from tests import integration


class NoSystemLibrariesTestCase(integration.TestCase):
    def test_system_libraries(self):
        self.run_snapcraft(["prime", "main-no-prereq"], "fake-curl-library")

        # Verify that the system's libcurl was pulled in.
        libcurl_path = os.path.join(
            self.prime_dir, "usr", "lib", self.arch_triplet, "libcurl.so*"
        )
        self.assertTrue(glob.glob(libcurl_path + "*"))

    def test_no_system_libraries(self):
        self.run_snapcraft(["prime", "main-no-libs"], "fake-curl-library")

        # Verify that the system's libcurl was NOT pulled in.
        self.assertThat(os.path.join(self.prime_dir, "usr"), Not(DirExists()))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools.matchers import Equals

from . import LifecycleCommandsBaseTestCase


class TestPack(LifecycleCommandsBaseTestCase):
    def test_using_defaults(self):
        result = self.run_command(["pack", "snap-dir"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_lifecycle_execute.mock.assert_not_called()
        self.fake_pack.mock.assert_called_once_with("snap-dir", output=None)

    def test_output(self):
        result = self.run_command(["pack", "snap-dir", "--output", "foo.snap"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_lifecycle_execute.mock.assert_not_called()
        self.fake_pack.mock.assert_called_once_with("snap-dir", output="foo.snap")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Operations with platform specific package repositories."""

from . import errors  # noqa
from . import snaps  # noqa
from . import _platform
from ._base import BaseRepo  # noqa
from ._base import fix_pkg_config  # noqa

# Imported for backwards compatibility with plugins
if _platform._is_deb_based():
    from ._deb import Ubuntu  # noqa

Repo = _platform._get_repo_for_platform()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import Optional
from unittest import mock

import fixtures
from testtools.matchers import Equals

import snapcraft.yaml_utils.errors
from snapcraft.internal import steps
from snapcraft.internal.build_providers.errors import ProviderExecError
from tests import fixture_setup
from tests.unit.build_providers import ProviderImpl

from . import CommandBaseTestCase


class LifecycleCommandsBaseTestCase(CommandBaseTestCase):

    yaml_template = """name: {step}-test
version: "1.0"
summary: test {step}
description: if the {step} is successful the state file will be updated
confinement: strict
grade: stable
base: {base}

parts:
{parts}"""

    yaml_part = """  {step}{iter:d}:
    plugin: nil"""

    def make_snapcraft_yaml(
        self, step, n=1, yaml_part=None, create=False, base="core18"
    ):
        if not yaml_part:
            yaml_part = self.yaml_part

        parts = "\n".join([yaml_part.format(step=step, iter=i) for i in range(n)])
        super().make_snapcraft_yaml(
            self.yaml_template.format(step=step, parts=parts, base=base)
        )

        parts = []
        for i in range(n):
            part_dir = os.path.join(self.parts_dir, "{}{}".format(step, i))
            state_dir = os.path.join(part_dir, "state")
            parts.append({"part_dir": part_dir, "state_dir": state_dir})

        return parts


class BuildEnvironmentParsingTest(LifecycleCommandsBaseTestCase):
    def setUp(self):
        super().setUp()

        # Use "clean" step for testing provider handling.
        self.step = "clean"

        # Don't actually run clean - we only want to test the command
        # line interface flag parsing.
        self.useFixture(fixtures.MockPatch("snapcraft.internal.lifecycle.clean"))

        # tests.unit.TestCase sets SNAPCRAFT_BUILD_ENVIRONMENT to host.
        # These build provider tests will want to set this explicitly.
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", None)
        )

        self.mock_get_provider_for = self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.build_providers.get_provider_for",
                return_value=ProviderImpl,
            )
        ).mock

        # Tests need to dictate this (or not).
        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.common.is_process_container", return_value=False
            )
        )

        self.useFixture(fixture_setup.FakeMultipass())

        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path, base="core")
        snapcraft_yaml.update_part("part1", dict(plugin="nil"))
        self.useFixture(snapcraft_yaml)


class AssortedBuildEnvironmentParsingTests(BuildEnvironmentParsingTest):
    def test_host_container(self):
        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.common.is_process_container", return_value=True
            )
        )
        result = self.run_command([self.step])

        # Mock get_provider_for not called when using "host".
        self.mock_get_provider_for.assert_not_called()
        self.assertThat(result.exit_code, Equals(0))

    def test_use_lxd(self):
        result = self.run_command([self.step, "--use-lxd"])

        self.mock_get_provider_for.assert_called_once_with("lxd")
        self.assertThat(result.exit_code, Equals(0))

    def test_destructive_mode(self):
        result = self.run_command([self.step, "--destructive-mode"])

        # Mock get_provider_for not called when using "host".
        self.mock_get_provider_for.assert_not_called()
        self.assertThat(result.exit_code, Equals(0))

    def test_host_build_env(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "host")
        )

        result = self.run_command([self.step])

        self.mock_get_provider_for.assert_not_called()
        self.assertThat(result.exit_code, Equals(0))

    def test_invalid_clean_provider_build_env(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "invalid")
        )

        result = self.run_command([self.step])

        self.assertThat(result.exit_code, Equals(2))

    def test_invalid_clean_provider_build_arg(self):
        result = self.run_command([self.step, "--provider", "invalid"])

        self.assertThat(result.exit_code, Equals(2))

    def test_invalid_combo_lxd_destructive(self):
        result = self.run_command([self.step, "--use-lxd", "--destructive-mode"])

        self.assertThat(result.exit_code, Equals(2))

    def test_invalid_combo_use_lxd_mismatch(self):
        result = self.run_command([self.step, "--use-lxd", "--provider", "host"])

        self.assertThat(result.exit_code, Equals(2))

    def test_invalid_combo_destructive_mode_mismatch(self):
        result = self.run_command(
            [self.step, "--destructive-mode", "--provider", "lxd"]
        )

        self.assertThat(result.exit_code, Equals(2))

    def test_invalid_host_without_destructive_mode(self):
        result = self.run_command([self.step, "--provider", "host"])

        self.assertThat(result.exit_code, Equals(2))

    def test_invalid_mismatch(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "multipass")
        )

        result = self.run_command([self.step, "--provider", "host"])

        self.assertThat(result.exit_code, Equals(2))


class ValidBuildEnvironmentParsingTests(BuildEnvironmentParsingTest):
    def assert_run(self, env, arg, result):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", env)
        )

        if result == "host":
            res = self.run_command([self.step, "--provider", arg, "--destructive-mode"])

            self.mock_get_provider_for.assert_not_called()
        else:
            res = self.run_command([self.step, "--provider", arg])

            self.mock_get_provider_for.assert_called_once_with(result)

        self.assertThat(res.exit_code, Equals(0))

    def test_default(self):
        self.assert_run(env=None, arg=None, result="multipass")

    def test_host_env(self):
        self.assert_run(env="host", arg=None, result="host")

    def test_host_arg(self):
        self.assert_run(env=None, arg="host", result="host")

    def test_host_both(self):
        self.assert_run(env="host", arg="host", result="host")

    def test_lxd_env(self):
        self.assert_run(env="lxd", arg=None, result="lxd")

    def test_lxd_arg(self):
        self.assert_run(env=None, arg="lxd", result="lxd")

    def test_lxd_both(self):
        self.assert_run(env="lxd", arg="lxd", result="lxd")

    def test_multipass_env(self):
        self.assert_run(env="multipass", arg=None, result="multipass")

    def test_multipass_arg(self):
        self.assert_run(env=None, arg="multipass", result="multipass")

    def test_multipass_both(self):
        self.assert_run(env="multipass", arg="multipass", result="multipass")


class BuildProviderYamlValidationTest(LifecycleCommandsBaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "multipass")
        )

        patcher = mock.patch(
            "snapcraft.internal.build_providers.get_provider_for",
            return_value=ProviderImpl,
        )
        self.provider = patcher.start()
        self.addCleanup(patcher.stop)

        self.useFixture(fixture_setup.FakeMultipass())

    def test_validation_passes(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path, base="core")
        snapcraft_yaml.update_part("part1", dict(plugin="nil"))
        self.useFixture(snapcraft_yaml)

        result = self.run_command(["pull"])

        self.assertThat(result.exit_code, Equals(0))

    def test_validation_fails(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path, name="name with spaces", base="core"
        )
        snapcraft_yaml.update_part("part1", dict(plugin="nil"))
        self.useFixture(snapcraft_yaml)

        self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError, self.run_command, ["pull"]
        )


class BuildProviderDebugCommandTestCase(LifecycleCommandsBaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "multipass")
        )
        self.useFixture(fixture_setup.FakeMultipass())

        shell_mock = mock.Mock()

        class Provider(ProviderImpl):
            def execute_step(self, step: steps.Step) -> None:
                raise ProviderExecError(
                    provider_name="fake", command=["snapcraft", "pull"], exit_code=1
                )

            def shell(self):
                shell_mock()

        patcher = mock.patch(
            "snapcraft.internal.build_providers.get_provider_for", return_value=Provider
        )
        self.provider = patcher.start()
        self.addCleanup(patcher.stop)

        self.shell_mock = shell_mock

        self.make_snapcraft_yaml("pull", base="core")

    def test_step_with_debug_using_build_provider_fails(self):
        result = self.run_command(["--debug", "pull"])

        self.assertThat(result.exit_code, Equals(0))
        self.shell_mock.assert_called_once_with()

    def test_step_with_debug_after_pull_using_build_provider_fails(self):
        result = self.run_command(["pull", "--debug"])

        self.assertThat(result.exit_code, Equals(0))
        self.shell_mock.assert_called_once_with()

    def test_step_without_debug_using_build_provider_fails_and_does_not_shell(self):
        self.assertRaises(ProviderExecError, self.run_command, ["pull"])

        self.shell_mock.assert_not_called()


class BuildProviderShellCommandTestCase(LifecycleCommandsBaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "multipass")
        )
        self.useFixture(fixture_setup.FakeMultipass())

        shell_mock = mock.Mock()
        pack_project_mock = mock.Mock()
        execute_step_mock = mock.Mock()

        class Provider(ProviderImpl):
            def pack_project(self, *, output: Optional[str] = None) -> None:
                pack_project_mock(output)

            def execute_step(self, step: steps.Step) -> None:
                execute_step_mock(step)

            def shell(self):
                shell_mock()

        patcher = mock.patch(
            "snapcraft.internal.build_providers.get_provider_for", return_value=Provider
        )
        self.provider = patcher.start()
        self.addCleanup(patcher.stop)

        self.shell_mock = shell_mock
        self.pack_project_mock = pack_project_mock
        self.execute_step_mock = execute_step_mock

        self.make_snapcraft_yaml("pull", base="core")

    def test_step_with_shell_after(self):
        result = self.run_command(["pull", "--shell-after"])

        self.assertThat(result.exit_code, Equals(0))
        self.pack_project_mock.assert_not_called()
        self.execute_step_mock.assert_called_once_with(steps.PULL)
        self.shell_mock.assert_called_once_with()

    def test_snap_with_shell_after(self):
        result = self.run_command(["snap", "--output", "fake.snap", "--shell-after"])

        self.assertThat(result.exit_code, Equals(0))
        self.pack_project_mock.assert_called_once_with("fake.snap")
        self.execute_step_mock.assert_not_called()
        self.shell_mock.assert_called_once_with()

    def test_step_without_shell_after_does_not_enter_shell(self):
        result = self.run_command(["pull"])

        self.assertThat(result.exit_code, Equals(0))
        self.pack_project_mock.assert_not_called()
        self.execute_step_mock.assert_called_once_with(steps.PULL)
        self.shell_mock.assert_not_called()

    def test_error_with_shell_after_error_and_debug(self):
        self.shell_mock.side_effect = EnvironmentError("error")

        self.assertRaises(
            EnvironmentError, self.run_command, ["pull", "--shell-after", "--debug"]
        )

        self.pack_project_mock.assert_not_called()
        self.execute_step_mock.assert_called_once_with(steps.PULL)
        self.shell_mock.assert_called_once_with()

    def test_pull_step_with_shell(self):
        result = self.run_command(["pull", "--shell"])

        self.assertThat(result.exit_code, Equals(0))
        self.pack_project_mock.assert_not_called()
        self.execute_step_mock.assert_not_called()
        self.shell_mock.assert_called_once_with()

    def test_step_with_shell(self):
        result = self.run_command(["stage", "--shell"])

        self.assertThat(result.exit_code, Equals(0))
        self.pack_project_mock.assert_not_called()
        self.execute_step_mock.assert_called_once_with(steps.BUILD)
        self.shell_mock.assert_called_once_with()

    def test_snap_with_shell(self):
        result = self.run_command(["snap", "--shell"])

        self.assertThat(result.exit_code, Equals(0))
        self.pack_project_mock.assert_not_called()
        self.execute_step_mock.assert_called_once_with(steps.PRIME)
        self.shell_mock.assert_called_once_with()

    def test_snap_without_shell(self):
        result = self.run_command(["snap"])

        self.assertThat(result.exit_code, Equals(0))
        self.pack_project_mock.assert_called_once_with(None)
        self.execute_step_mock.assert_not_called()
        self.shell_mock.assert_not_called()


class BuildProviderTryCommandTestCase(LifecycleCommandsBaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "multipass")
        )
        self.useFixture(fixture_setup.FakeMultipass())

        mount_prime_mock = mock.Mock(return_value=False)

        class Provider(ProviderImpl):
            def _mount_prime_directory(self) -> bool:
                return mount_prime_mock()

        patcher = mock.patch(
            "snapcraft.internal.build_providers.get_provider_for", return_value=Provider
        )
        self.provider = patcher.start()
        self.addCleanup(patcher.stop)

        self.mount_prime_mock = mount_prime_mock

        self.make_snapcraft_yaml("pull", base="core")

    def test_try(self):
        result = self.run_command(["try"])

        self.assertThat(result.exit_code, Equals(0))
        self.mount_prime_mock.assert_called_once_with()


class BuildProviderCleanCommandTestCase(LifecycleCommandsBaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "multipass")
        )
        self.useFixture(fixture_setup.FakeMultipass())

        clean_project_mock = mock.Mock()
        clean_mock = mock.Mock()

        class Provider(ProviderImpl):
            def execute_step(self, step: steps.Step) -> None:
                raise ProviderExecError(
                    provider_name="fake", command=["snapcraft", "pull"], exit_code=1
                )

            def clean_project(self):
                clean_project_mock()

            def clean_parts(self, part_names):
                clean_mock(part_names=part_names)

        patcher = mock.patch(
            "snapcraft.internal.build_providers.get_provider_for", return_value=Provider
        )
        self.provider = patcher.start()
        self.addCleanup(patcher.stop)

        self.clean_project_mock = clean_project_mock
        self.clean_mock = clean_mock

        self.make_snapcraft_yaml("pull", base="core")

    @mock.patch("snapcraft.internal.lifecycle.clean")
    def test_clean(self, lifecycle_clean_mock):
        result = self.run_command(["clean"])

        self.assertThat(result.exit_code, Equals(0))
        lifecycle_clean_mock.assert_called_once_with(mock.ANY, tuple(), steps.PRIME)
        self.clean_project_mock.assert_called_once_with()
        self.clean_mock.assert_not_called()

    def test_clean_a_single_part(self):
        result = self.run_command(["clean", "part1"])

        self.assertThat(result.exit_code, Equals(0))
        self.clean_project_mock.assert_not_called()
        self.clean_mock.assert_called_once_with(part_names=("part1",))

    def test_clean_multiple_parts(self):
        result = self.run_command(["clean", "part1", "part2", "part3"])

        self.assertThat(result.exit_code, Equals(0))
        self.clean_project_mock.assert_not_called()
        self.clean_mock.assert_called_once_with(part_names=("part1", "part2", "part3"))

    def test_unprime_with_build_environment_errors(self):
        result = self.run_command(["clean", "--unprime"])

        self.assertThat(result.exit_code, Equals(2))
        self.clean_project_mock.assert_not_called()
        self.clean_mock.assert_not_called()

    @mock.patch("snapcraft.cli.lifecycle.get_project")
    @mock.patch("snapcraft.internal.lifecycle.clean")
    def test_unprime_in_managed_host(self, lifecycle_clean_mock, get_project_mock):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "managed-host")
        )
        project_mock = mock.Mock()
        project_mock.info.base.return_value = "core"
        get_project_mock.return_value = project_mock

        result = self.run_command(["clean", "--unprime"])

        self.assertThat(result.exit_code, Equals(0))
        lifecycle_clean_mock.assert_called_once_with(project_mock, tuple(), steps.PRIME)
        self.clean_project_mock.assert_not_called()
        self.clean_mock.assert_not_called()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

import pytest
from testtools.matchers import Equals, Is, MatchesRegex

import snapcraft.yaml_utils.errors
from snapcraft.project._project_info import ProjectInfo
from tests import unit


class ProjectInfoTest(unit.TestCase):
    def test_properties(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            dedent(
                """\
            name: foo
            version: "1"
            summary: bar
            description: baz
            confinement: strict
        """
            )
        )

        info = ProjectInfo(snapcraft_yaml_file_path=snapcraft_yaml_file_path)
        self.assertThat(info.name, Equals("foo"))
        self.assertThat(info.version, Equals("1"))
        self.assertThat(info.summary, Equals("bar"))
        self.assertThat(info.description, Equals("baz"))
        self.assertThat(info.confinement, Equals("strict"))

    def test_empty_yaml(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml("")

        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            ProjectInfo,
            snapcraft_yaml_file_path=snapcraft_yaml_file_path,
        )

        self.assertThat(
            raised.message,
            Equals(
                "'name' is a required property in {!r}".format(snapcraft_yaml_file_path)
            ),
        )

    def test_minimal_load_with_name_only(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            dedent(
                """\
            name: foo
        """
            )
        )

        info = ProjectInfo(snapcraft_yaml_file_path=snapcraft_yaml_file_path)
        self.assertThat(info.name, Equals("foo"))
        self.assertThat(info.version, Is(None))
        self.assertThat(info.summary, Is(None))
        self.assertThat(info.description, Is(None))
        self.assertThat(info.confinement, Is(None))

    def test_name_is_required(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            dedent(
                """\
            version: "1"
            summary: bar
            description: baz
            confinement: strict
        """
            )
        )

        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            ProjectInfo,
            snapcraft_yaml_file_path=snapcraft_yaml_file_path,
        )

        self.assertThat(
            raised.message,
            Equals(
                "'name' is a required property in {!r}".format(snapcraft_yaml_file_path)
            ),
        )

    def test_get_raw_dict(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            dedent(
                """\
            name: foo
            version: "1"
            summary: bar
            description: baz
            confinement: strict
        """
            )
        )

        info = ProjectInfo(snapcraft_yaml_file_path=snapcraft_yaml_file_path)
        raw_snapcraft = info.get_raw_snapcraft()

        self.assertThat(raw_snapcraft["name"], Equals("foo"))
        self.assertThat(raw_snapcraft["version"], Equals("1"))
        self.assertThat(raw_snapcraft["summary"], Equals("bar"))
        self.assertThat(raw_snapcraft["description"], Equals("baz"))
        self.assertThat(raw_snapcraft["confinement"], Equals("strict"))

    def test_get_raw_dict_is_a_copy(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            dedent(
                """\
            name: foo
            version: "1"
            summary: bar
            description: baz
            confinement: strict
        """
            )
        )

        info = ProjectInfo(snapcraft_yaml_file_path=snapcraft_yaml_file_path)
        raw_snapcraft = info.get_raw_snapcraft()
        raw_snapcraft.pop("name")
        self.assertThat(raw_snapcraft.get("name"), Is(None))

        raw_snapcraft = info.get_raw_snapcraft()
        self.assertThat(raw_snapcraft.get("name"), Equals("foo"))


class InvalidYamlTest(unit.TestCase):
    def test_tab_in_yaml(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            dedent(
                """\
            name: test
            version: "1"
            summary: test
            description: nothing
            \tconfinement: strict
            grade: stable

            parts:
              part1:
                plugin: go
                stage-packages: [fswebcam]
        """
            )
        )

        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            ProjectInfo,
            snapcraft_yaml_file_path=snapcraft_yaml_file_path,
        )

        self.assertThat(raised.source, Equals(snapcraft_yaml_file_path))
        # libyaml had a spelling mistake indentation/intendation
        self.assertThat(
            raised.message,
            MatchesRegex(
                "found a tab character that violate (indentation|intendation)"
                " on line 5, column 1"
            ),
        )

    def test_invalid_yaml_invalid_unicode_chars(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            dedent(
                """\
            name: foobar
            version: "1"
            summary: test\uffff
            description: nothing
        """
            )
        )

        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            ProjectInfo,
            snapcraft_yaml_file_path=snapcraft_yaml_file_path,
        )

        self.assertThat(raised.source, Equals(snapcraft_yaml_file_path))
        self.assertThat(
            raised.message,
            Equals(
                "invalid character '\\uffff' at position 40: control characters are not allowed"
            ),
        )

    def test_invalid_yaml_unhashable(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            dedent(
                """\
                name: test
                version: {{invalid}}
                summary: test
                description: test
                confinement: strict
                grade: stable
                parts:
                part1:
                    plugin: nil
                """
            )
        )

        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            ProjectInfo,
            snapcraft_yaml_file_path=snapcraft_yaml_file_path,
        )

        self.assertThat(raised.source, Equals(snapcraft_yaml_file_path))
        self.assertThat(
            raised.message, Equals("found unhashable key on line 2, column 10")
        )

    def test_invalid_yaml_list_in_mapping(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            dedent(
                """\
                name: foobar
                - list item
                """
            )
        )

        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            ProjectInfo,
            snapcraft_yaml_file_path=snapcraft_yaml_file_path,
        )

        self.assertThat(raised.source, Equals(snapcraft_yaml_file_path))
        self.assertThat(
            raised.message, Equals("did not find expected key on line 2, column 1")
        )


@pytest.mark.parametrize("encoding", ["utf-8", "utf-8-sig", "utf-16"])
def test_different_encodings_loads(tmp_work_path, encoding):
    snapcraft_yaml = dedent(
        """\
        name: test
        version: "1"
        summary: test
        description: ñoño test
        confinement: strict
        grade: stable

        parts:
          part1:
            plugin: go
            stage-packages: [fswebcam]
    """
    )

    snapcraft_yaml_path = tmp_work_path / "snapcraft.yaml"
    with snapcraft_yaml_path.open("w", encoding=encoding) as snapcraft_file:
        print(snapcraft_yaml, file=snapcraft_file)

    ProjectInfo(snapcraft_yaml_file_path=snapcraft_yaml_path.as_posix())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import os
import subprocess
from textwrap import dedent
from unittest import mock

from testtools.matchers import Contains, FileContains, FileExists

from snapcraft.internal import errors
from snapcraft.internal.pluginhandler import _runner
from tests import fixture_setup, unit


def _fake_pull():
    open(os.path.join("sourcedir", "fake-pull"), "w").close()


def _fake_build():
    open(os.path.join("builddir", "fake-build"), "w").close()


def _fake_stage():
    open(os.path.join("stagedir", "fake-stage"), "w").close()


def _fake_prime():
    open(os.path.join("primedir", "fake-prime"), "w").close()


class RunnerTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.partdir = os.path.abspath("partdir")
        os.mkdir(self.partdir)

    def test_pull(self):
        os.mkdir("sourcedir")

        runner = _runner.Runner(
            part_properties={"override-pull": "touch pull"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={},
            env_generator=lambda step: "export FOO=BAR",
        )

        runner.pull()

        self.assertThat(os.path.join("sourcedir", "pull"), FileExists())

    def test_builtin_function_from_pull(self):
        os.mkdir("sourcedir")

        runner = _runner.Runner(
            part_properties={"override-pull": "snapcraftctl pull"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={"pull": _fake_pull},
            env_generator=lambda step: "export FOO=BAR",
        )

        runner.pull()

        self.assertThat(os.path.join("sourcedir", "fake-pull"), FileExists())

    def test_snapcraft_utils_in_path_if_snap(self):
        self.useFixture(fixture_setup.FakeSnapcraftIsASnap())

        os.mkdir("builddir")

        runner = _runner.Runner(
            part_properties={"override-build": "echo $PATH > path"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={},
            env_generator=lambda step: "export FOO=BAR",
        )

        runner.build()

        expected_path_segment = "/snap/snapcraft/current/bin/scriptlet-bin"

        self.assertThat(os.path.join("builddir", "path"), FileExists())
        self.assertThat(
            os.path.join("builddir", "path"),
            FileContains(matcher=Contains(expected_path_segment)),
        )

    def test_build(self):
        os.mkdir("builddir")

        runner = _runner.Runner(
            part_properties={"override-build": "touch build"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={},
            env_generator=lambda step: "export FOO=BAR",
        )

        runner.build()

        self.assertThat(os.path.join("builddir", "build"), FileExists())

    def test_builtin_function_from_build(self):
        os.mkdir("builddir")

        runner = _runner.Runner(
            part_properties={"override-build": "snapcraftctl build"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={"build": _fake_build},
            env_generator=lambda step: "export FOO=BAR",
        )

        runner.build()

        self.assertThat(os.path.join("builddir", "fake-build"), FileExists())

    def test_stage(self):
        os.mkdir("stagedir")

        runner = _runner.Runner(
            part_properties={"override-stage": "touch stage"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={},
            env_generator=lambda step: "export FOO=BAR",
        )

        runner.stage()

        self.assertThat(os.path.join("stagedir", "stage"), FileExists())

    def test_builtin_function_from_stage(self):
        os.mkdir("stagedir")

        runner = _runner.Runner(
            part_properties={"override-stage": "snapcraftctl stage"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={"stage": _fake_stage},
            env_generator=lambda step: "export FOO=BAR",
        )

        runner.stage()

        self.assertThat(os.path.join("stagedir", "fake-stage"), FileExists())

    def test_prime(self):
        os.mkdir("primedir")

        runner = _runner.Runner(
            part_properties={"override-prime": "touch prime"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={},
            env_generator=lambda step: "export FOO=BAR",
        )

        runner.prime()

        self.assertThat(os.path.join("primedir", "prime"), FileExists())

    def test_builtin_function_from_prime(self):
        os.mkdir("primedir")

        runner = _runner.Runner(
            part_properties={"override-prime": "snapcraftctl prime"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={"prime": _fake_prime},
            env_generator=lambda step: "export FOO=BAR",
        )

        runner.prime()

        self.assertThat(os.path.join("primedir", "fake-prime"), FileExists())


class RunnerFailureTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.partdir = os.path.abspath("partdir")
        os.mkdir(self.partdir)

    def test_failure_on_last_script_command_results_in_failure(self):
        os.mkdir("builddir")

        script = dedent(
            """\
            touch success
            false  # this should trigger an error
        """
        )

        runner = _runner.Runner(
            part_properties={"override-build": script},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={},
            env_generator=lambda step: "export FOO=BAR",
        )

        self.assertRaises(errors.ScriptletRunError, runner.build)

    def test_failure_to_execute_mid_script_results_in_failure(self):
        os.mkdir("builddir")

        script = dedent(
            """\
            false  # this should trigger an error
            touch success
        """
        )

        runner = _runner.Runner(
            part_properties={"override-build": script},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={},
            env_generator=lambda step: "export FOO=BAR",
        )

        self.assertRaises(errors.ScriptletRunError, runner.build)

    def test_snapcraftctl_no_alias_if_not_snap(self):
        os.mkdir("builddir")

        runner = _runner.Runner(
            part_properties={"override-build": "alias snapcraftctl 2> /dev/null"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={},
            env_generator=lambda step: "export FOO=BAR",
        )

        self.assertRaises(errors.ScriptletRunError, runner.build)

    def test_snapcraftctl_errors_on_exception(self):
        os.mkdir("primedir")

        class _TestException(errors.ScriptletBaseError):
            fmt = "I'm an error"

        def _raise():
            raise _TestException()

        runner = _runner.Runner(
            part_properties={"override-prime": "snapcraftctl prime"},
            partdir=self.partdir,
            sourcedir="sourcedir",
            builddir="builddir",
            stagedir="stagedir",
            primedir="primedir",
            builtin_functions={"prime": _raise},
            env_generator=lambda step: "export FOO=BAR",
        )

        silent_popen = functools.partial(
            subprocess.Popen, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )

        with mock.patch("subprocess.Popen", wraps=silent_popen):
            self.assertRaises(_TestException, runner.prime)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testscenarios import multiply_scenarios

from snapcraft.internal import steps
from snapcraft.internal.pluginhandler._dirty_report import Dependency, DirtyReport


class TestDirtyReportGetReport:

    property_scenarios = [
        ("no properties", dict(dirty_properties=None, properties_report="")),
        (
            "single property",
            dict(
                dirty_properties=["prop1"],
                properties_report="The 'prop1' part property appears to have "
                "changed.",
            ),
        ),
        (
            "multiple properties",
            dict(
                dirty_properties=["prop1", "prop2"],
                properties_report="The 'prop1' and 'prop2' part properties appear "
                "to have changed.",
            ),
        ),
    ]

    option_scenarios = [
        ("no project options", dict(dirty_project_options=None, options_report="")),
        (
            "single project option",
            dict(
                dirty_project_options=["op1"],
                options_report="The 'op1' project option appears to have changed.",
            ),
        ),
        (
            "multiple properties",
            dict(
                dirty_project_options=["op1", "op2"],
                options_report="The 'op1' and 'op2' project options appear to "
                "have changed.",
            ),
        ),
    ]

    dependencies_scenarios = [
        (
            "no changed dependencies",
            dict(changed_dependencies=None, dependencies_report=""),
        ),
        (
            "single changed dependency",
            dict(
                changed_dependencies=[Dependency(part_name="dep1", step=steps.PULL)],
                dependencies_report="A dependency has changed: 'dep1'",
            ),
        ),
        (
            "multiple changed dependency",
            dict(
                changed_dependencies=[
                    Dependency(part_name="dep1", step=steps.PULL),
                    Dependency(part_name="dep2", step=steps.PULL),
                ],
                dependencies_report="Some dependencies have changed: 'dep1' and "
                "'dep2'",
            ),
        ),
    ]

    scenarios = multiply_scenarios(
        property_scenarios, option_scenarios, dependencies_scenarios
    )

    def test_get_report(
        self,
        dirty_properties,
        properties_report,
        dirty_project_options,
        options_report,
        changed_dependencies,
        dependencies_report,
    ):
        dirty_report = DirtyReport(
            dirty_properties=dirty_properties,
            dirty_project_options=dirty_project_options,
            changed_dependencies=changed_dependencies,
        )

        expected_report = []
        if properties_report:
            expected_report.append(properties_report)
        if options_report:
            expected_report.append(options_report)
        if dependencies_report:
            expected_report.append(dependencies_report)
        if expected_report:
            expected_report.append("")

        assert dirty_report.get_report() == "\n".join(expected_report)


class TestDirtyReportGetSummary:

    scenarios = [
        (
            "single property",
            dict(
                dirty_properties=["foo"],
                dirty_project_options=None,
                changed_dependencies=None,
                expected_summary="'foo' property changed",
            ),
        ),
        (
            "multiple properties",
            dict(
                dirty_properties=["foo", "bar"],
                dirty_project_options=None,
                changed_dependencies=None,
                expected_summary="properties changed",
            ),
        ),
        (
            "single option",
            dict(
                dirty_properties=None,
                dirty_project_options=["foo"],
                changed_dependencies=None,
                expected_summary="'foo' option changed",
            ),
        ),
        (
            "multiple options",
            dict(
                dirty_properties=None,
                dirty_project_options=["foo", "bar"],
                changed_dependencies=None,
                expected_summary="options changed",
            ),
        ),
        (
            "single dependency",
            dict(
                dirty_properties=None,
                dirty_project_options=None,
                changed_dependencies=[Dependency(part_name="foo", step=steps.PULL)],
                expected_summary="'foo' changed",
            ),
        ),
        (
            "multiple dependencies",
            dict(
                dirty_properties=None,
                dirty_project_options=None,
                changed_dependencies=[
                    Dependency(part_name="foo", step=steps.PULL),
                    Dependency(part_name="bar", step=steps.PULL),
                ],
                expected_summary="dependencies changed",
            ),
        ),
        (
            "property and option",
            dict(
                dirty_properties=["foo"],
                dirty_project_options=["bar"],
                changed_dependencies=None,
                expected_summary="options and properties changed",
            ),
        ),
        (
            "property and dependencies",
            dict(
                dirty_properties=["foo"],
                dirty_project_options=None,
                changed_dependencies=[Dependency(part_name="bar", step=steps.PULL)],
                expected_summary="dependencies and properties changed",
            ),
        ),
        (
            "options and dependencies",
            dict(
                dirty_properties=None,
                dirty_project_options=["foo"],
                changed_dependencies=[Dependency(part_name="bar", step=steps.PULL)],
                expected_summary="dependencies and options changed",
            ),
        ),
    ]

    def test_get_summary(
        self,
        dirty_properties,
        dirty_project_options,
        changed_dependencies,
        expected_summary,
    ):
        dirty_report = DirtyReport(
            dirty_properties=dirty_properties,
            dirty_project_options=dirty_project_options,
            changed_dependencies=changed_dependencies,
        )

        assert dirty_report.get_summary() == expected_summary



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import fixtures
from click.testing import CliRunner

from snapcraft.cli.snapcraftctl._runner import run
from tests import unit


class CommandBaseNoFifoTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.runner = CliRunner()

    def run_command(self, args, **kwargs):
        return self.runner.invoke(run, args, catch_exceptions=False, **kwargs)


class CommandBaseTestCase(CommandBaseNoFifoTestCase):
    def setUp(self):
        super().setUp()

        tempdir = self.useFixture(fixtures.TempDir()).path
        self.call_fifo = os.path.join(tempdir, "call_fifo")
        self.feedback_fifo = os.path.join(tempdir, "feedback_fifo")

        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFTCTL_CALL_FIFO", self.call_fifo)
        )
        self.useFixture(
            fixtures.EnvironmentVariable(
                "SNAPCRAFTCTL_FEEDBACK_FIFO", self.feedback_fifo
            )
        )

        # Create the FIFOs
        open(self.call_fifo, "w").close()
        open(self.feedback_fifo, "w").close()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib

import pymacaroons
import pytest

from snapcraft.storeapi import http_clients


def test_invalid_macaroon_root_raises_exception(tmp_work_path):
    with pathlib.Path("conf").open("w") as config_fd:
        print("[login.ubuntu.com]", file=config_fd)
        print("macaroon=inval'id", file=config_fd)
        config_fd.flush()

    client = http_clients.UbuntuOneAuthClient()
    with pathlib.Path("conf").open() as config_fd:
        with pytest.raises(http_clients.errors.InvalidCredentialsError):
            client.login(config_fd=config_fd)


def test_invalid_discharge_raises_exception():
    with pathlib.Path("conf").open("w") as config_fd:
        print("[login.ubuntu.com]", file=config_fd)
        print("macaroon={}".format(pymacaroons.Macaroon().serialize()), file=config_fd)
        print("unbound_discharge=inval'id", file=config_fd)
        config_fd.flush()

    client = http_clients.UbuntuOneAuthClient()

    with pathlib.Path("conf").open() as config_fd:
        with pytest.raises(http_clients.errors.InvalidCredentialsError):
            client.login(config_fd=config_fd)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gzip
import logging
import os
import shutil
import time
from datetime import datetime
from typing import Any, Dict, List, Optional, Sequence, Type
from urllib.parse import unquote, urlsplit

import requests
from launchpadlib.launchpad import Launchpad
from lazr import restfulclient
from lazr.restfulclient.resource import Entry
from xdg import BaseDirectory

import snapcraft
from snapcraft.internal.sources._git import Git
from snapcraft.internal.sources.errors import SnapcraftPullError
from snapcraft.project import Project

from . import errors

_LP_POLL_INTERVAL = 30
_LP_SUCCESS_STATUS = "Successfully built"
_LP_FAIL_STATUS = "Failed to build"

logger = logging.getLogger(__name__)


def _is_build_pending(build: Dict[str, Any]) -> bool:
    # Possible values according to API documentation:
    # - "Needs building"
    # - "Successfully built"
    # - "Failed to build"
    # - "Dependency wait"
    # - "Chroot problem"
    # - "Build for superseded Source"
    # - "Currently building"
    # - "Failed to upload"
    # - "Uploading build"
    # - "Cancelling build"
    # - "Cancelled build"
    if _is_build_status_success(build) or _is_build_status_failure(build):
        return False

    return True


def _is_build_status_success(build: Dict[str, Any]) -> bool:
    build_state = build["buildstate"]
    return build_state == "Successfully built"


def _is_build_status_failure(build: Dict[str, Any]) -> bool:
    build_state = build["buildstate"]
    return build_state in ["Failed to build", "Cancelled build"]


def _get_url_basename(url: str):
    path = urlsplit(url).path
    return unquote(path).split("/")[-1]


class LaunchpadClient:
    """Launchpad remote builder operations."""

    def __init__(
        self,
        *,
        project: Project,
        build_id: str,
        architectures: Sequence[str],
        git_branch: str = "master",
        core18_channel: str = "stable",
        snapcraft_channel: str = "stable",
        deadline: int = 0,
        git_class: Type[Git] = Git,
        running_snapcraft_version: str = snapcraft.__version__,
    ) -> None:
        self._git_class = git_class
        if not self._git_class.check_command_installed():
            raise errors.GitNotFoundProviderError(provider="Launchpad")

        self._snap_name = project.info.name
        self._build_id = build_id

        self.architectures = architectures

        self._lp_name = build_id
        self._lp_git_branch = git_branch

        self._core18_channel = core18_channel
        self._snapcraft_channel = snapcraft_channel
        self._running_snapcraft_version = running_snapcraft_version

        self._cache_dir = self._create_cache_directory()
        self._data_dir = self._create_data_directory()
        self._credentials = os.path.join(self._data_dir, "credentials")

        self._lp: Launchpad = self.login()
        self.user = self._lp.me.name

        self.deadline = deadline

    @property
    def architectures(self) -> Sequence[str]:
        return self._architectures

    @architectures.setter
    def architectures(self, architectures: Sequence[str]) -> None:
        self._lp_processors: Optional[Sequence[str]] = None

        if architectures:
            self._lp_processors = ["/+processors/" + a for a in architectures]

        self._architectures = architectures

    @property
    def user(self) -> str:
        return self._lp_user

    @user.setter
    def user(self, user: str) -> None:
        self._lp_user = user
        self._lp_owner = f"/~{user}"

    def _check_timeout_deadline(self) -> None:
        if self.deadline <= 0:
            return

        if int(time.time()) >= self.deadline:
            raise errors.RemoteBuildTimeoutError()

    def _create_data_directory(self) -> str:
        data_dir = BaseDirectory.save_data_path("snapcraft", "provider", "launchpad")
        os.makedirs(data_dir, mode=0o700, exist_ok=True)
        return data_dir

    def _create_cache_directory(self) -> str:
        cache_dir = BaseDirectory.save_cache_path("snapcraft", "provider", "launchpad")
        os.makedirs(cache_dir, mode=0o700, exist_ok=True)
        return cache_dir

    def _fetch_artifacts(self, snap: Entry) -> None:
        """Fetch build arftifacts (logs and snaps)."""
        builds = self._get_builds(snap)

        logger.debug("Downloading artifacts...")
        for build in builds:
            self._download_build_artifacts(build)
            self._download_log(build)

    def _get_builds_collection_entry(self, snap: Entry) -> Optional[Entry]:
        logger.debug("Fetching builds collection information from Launchpad...")
        url = snap.builds_collection_link
        return self._lp_load_url(url)

    def _get_builds(self, snap: Entry) -> List[Dict[str, Any]]:
        bc = self._get_builds_collection_entry(snap)
        if bc is None:
            return []

        return bc.entries

    def _get_snap(self) -> Optional[Entry]:
        try:
            return self._lp.snaps.getByName(name=self._lp_name, owner=self._lp_owner)
        except restfulclient.errors.NotFound:
            return None

    def _issue_build_request(self, snap: Entry) -> Entry:
        dist = self._lp.distributions["ubuntu"]
        archive = dist.main_archive
        return snap.requestBuilds(
            archive=archive,
            channels={
                "core18": self._core18_channel,
                "snapcraft": self._snapcraft_channel,
            },
            pocket="Updates",
        )

    def _lp_load_url(self, url: str) -> Entry:
        """Load Launchpad url with a retry in case the connection is lost."""
        try:
            return self._lp.load(url)
        except ConnectionResetError:
            self._lp = self.login()
            return self._lp.load(url)

    def _wait_for_build_request_acceptance(
        self, build_request: Entry, timeout: int = 0
    ) -> None:
        # Not be be confused with the actual build(s), this is
        # ensuring that Launchpad accepts the build request.
        while build_request.status == "Pending":
            # Check to see if we've run out of time.
            self._check_timeout_deadline()

            logger.debug("Waiting on Launchpad build request...")
            logger.debug(
                f"status={build_request.status} error={build_request.error_message}"
            )

            time.sleep(1)

            # Refresh status.
            build_request.lp_refresh()

        if build_request.status == "Failed":
            # Build request failed.
            self.cleanup()
            raise errors.RemoteBuilderError(builder_error=build_request.error_message)
        elif build_request.status != "Completed":
            # Shouldn't end up here.
            self.cleanup()
            raise errors.RemoteBuilderError(
                builder_error="Unknown builder error - reported status: {}".format(
                    build_request.status
                )
            )
        elif not build_request.builds.entries:
            # Shouldn't end up here either.
            self.cleanup()
            raise errors.RemoteBuilderError(
                builder_error="Unknown builder error - no build entries found."
            )

        build_number = _get_url_basename(build_request.self_link)
        logger.debug(f"Build request accepted: {build_number}")

    def login(self) -> Launchpad:
        try:
            return Launchpad.login_with(
                "snapcraft remote-build {}".format(snapcraft.__version__),
                "production",
                self._cache_dir,
                credentials_file=self._credentials,
                version="devel",
            )
        except (ConnectionRefusedError, TimeoutError):
            raise errors.LaunchpadHttpsError()

    def get_git_repo_path(self) -> str:
        return f"~{self._lp_user}/+git/{self._lp_name}"

    def get_git_https_url(self, token: Optional[str] = None) -> str:
        if token:
            return f"https://{self._lp_user}:{token}@git.launchpad.net/~{self._lp_user}/+git/{self._lp_name}/"
        else:
            return f"https://{self._lp_user}@git.launchpad.net/~{self._lp_user}/+git/{self._lp_name}/"

    def _create_git_repository(self, force=False) -> Entry:
        """Create git repository."""
        if force:
            self._delete_git_repository()

        logger.debug(
            f"creating git repo: name={self._lp_name}, owner={self._lp_owner}, target={self._lp_owner}"
        )
        return self._lp.git_repositories.new(
            name=self._lp_name, owner=self._lp_owner, target=self._lp_owner
        )

    def _delete_git_repository(self) -> None:
        """Delete git repository."""
        git_path = self.get_git_repo_path()
        git_repo = self._lp.git_repositories.getByPath(path=git_path)

        # git_repositories.getByPath returns None if git repo does not exist.
        if git_repo is None:
            return

        logger.debug("Deleting source repository from Launchpad...")
        git_repo.lp_delete()

    def _create_snap(self, force=False) -> Entry:
        """Create a snap recipe. Use force=true to replace existing snap."""
        git_url = self.get_git_https_url()

        if force:
            self._delete_snap()

        optional_kwargs = dict()
        if self._lp_processors:
            optional_kwargs["processors"] = self._lp_processors

        logger.debug("Registering snap job on Launchpad...")
        logger.debug(f"url=https://launchpad.net{self._lp_owner}/+snap/{self._lp_name}")

        return self._lp.snaps.new(
            name=self._lp_name,
            owner=self._lp_owner,
            git_repository_url=git_url,
            git_path=self._lp_git_branch,
            auto_build=False,
            auto_build_archive="/ubuntu/+archive/primary",
            auto_build_pocket="Updates",
            **optional_kwargs,
        )

    def _delete_snap(self) -> None:
        """Remove snap info and all associated files."""
        snap = self._get_snap()
        if snap is None:
            return

        logger.debug("Removing snap job from Launchpad...")
        snap.lp_delete()

    def cleanup(self) -> None:
        """Delete snap and git repository from launchpad."""
        self._delete_snap()
        self._delete_git_repository()

    def start_build(self, timeout: int = 0) -> None:
        """Start build with specified timeout (time.time() in seconds)."""
        snap = self._create_snap(force=True)

        logger.debug("Issuing build request on Launchpad...")
        build_request = self._issue_build_request(snap)
        self._wait_for_build_request_acceptance(build_request, timeout=timeout)

    def monitor_build(
        self, interval: int = _LP_POLL_INTERVAL, timeout: int = 0
    ) -> None:
        """Check build progress, and download artifacts when ready."""
        snap = self._get_snap()

        while True:
            # Check to see if we've run out of time.
            self._check_timeout_deadline()

            builds = self._get_builds(snap)
            pending = False
            timestamp = str(datetime.now())
            logger.info(f"Build status as of {timestamp}:")
            for build in builds:
                state = build["buildstate"]
                arch = build["arch_tag"]
                logger.info(f"\tarch={arch}\tstate={state}")

                if _is_build_pending(build):
                    pending = True

            if pending is False:
                break

            time.sleep(interval)

        # Build is complete - download build artifacts.
        self._fetch_artifacts(snap)

    def get_build_status(self) -> Dict[str, str]:
        """Get status of builds."""
        snap = self._get_snap()
        builds = self._get_builds(snap)
        build_status: Dict[str, str] = dict()
        for build in builds:
            state = build["buildstate"]
            arch = build["arch_tag"]
            build_status[arch] = state

        return build_status

    def _get_logfile_name(self, arch: str) -> str:
        n = 0
        base_name = "{}_{}".format(self._snap_name, arch)
        log_name = f"{base_name}.txt"

        while os.path.isfile(log_name):
            n += 1
            log_name = f"{base_name}.{n}.txt"

        return log_name

    def _download_log(self, build: Dict[str, Any]) -> None:
        url = build["build_log_url"]
        arch = build["arch_tag"]
        if url is None:
            logger.info(f"No build log available for {arch!r}.")
        else:
            log_name = self._get_logfile_name(arch)
            self._download_file(url=url, dst=log_name, gunzip=True)
            logger.info(f"Build log available at {log_name!r}")

        if _is_build_status_failure(build):
            logger.error(f"Build failed for arch {arch!r}.")

    def _download_file(self, *, url: str, dst: str, gunzip: bool = False) -> None:
        # TODO: consolidate with, and use indicators.download_requests_stream
        logger.debug(f"Downloading: {url}")
        try:
            with requests.get(url, stream=True) as response:
                # Wrap response with gzipfile if gunzip is requested.
                stream = response.raw
                if gunzip:
                    stream = gzip.GzipFile(fileobj=stream)
                with open(dst, "wb") as f_dst:
                    shutil.copyfileobj(stream, f_dst)
                response.raise_for_status()
        except requests.exceptions.RequestException as e:
            logger.error(f"Error downloading {url}: {str(e)}")

    def _download_build_artifacts(self, build: Dict[str, Any]) -> None:
        arch = build["arch_tag"]
        snap_build = self._lp_load_url(build["self_link"])
        urls = snap_build.getFileUrls()

        if not urls:
            logger.error(f"Snap file not available for arch {arch!r}.")
            return

        for url in urls:
            file_name = _get_url_basename(url)

            self._download_file(url=url, dst=file_name)

            if file_name.endswith(".snap"):
                logger.info(f"Snapped {file_name}")
            else:
                logger.info(f"Fetched {file_name}")

    def _gitify_repository(self, repo_dir: str) -> Git:
        """Git-ify source repository tree.

        :return: Git handler instance to git repository.
        """
        git_handler = self._git_class(repo_dir, repo_dir, silent=True)

        # Init repo.
        git_handler.init()

        # Add all files found in repo.
        for f in os.listdir(repo_dir):
            if f != ".git":
                git_handler.add(f)

        # Commit files.
        git_handler.commit(
            f"committed by snapcraft version: {self._running_snapcraft_version}"
        )

        return git_handler

    def has_outstanding_build(self) -> bool:
        """Check if there is an existing build configured on Launchpad."""
        snap = self._get_snap()
        return snap is not None

    def push_source_tree(self, repo_dir: str) -> str:
        """Push source tree to Launchpad, returning URL."""
        git_handler = self._gitify_repository(repo_dir)
        lp_repo = self._create_git_repository(force=True)
        token = lp_repo.issueAccessToken()

        url = self.get_git_https_url(token=token)
        stripped_url = self.get_git_https_url(token="<token>")

        logger.info(f"Sending build data to Launchpad... ({stripped_url})")

        try:
            git_handler.push(url, f"HEAD:{self._lp_git_branch}", force=True)
        except SnapcraftPullError as error:
            # Strip token from command.
            command = error.command.replace(token, "<token>")  # type: ignore
            exit_code = error.exit_code  # type: ignore
            raise errors.LaunchpadGitPushError(command=command, exit_code=exit_code)

        return url



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import fixtures
from testtools.matchers import Equals

from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase, get_sample_key


class CreateKeyTestCase(FakeStoreCommandsBaseTestCase):
    def setUp(self):
        super().setUp()

        self.fake_check_call = fixtures.MockPatch("subprocess.check_call")
        self.useFixture(self.fake_check_call)

    def test_create_key_already_exists(self):
        raised = self.assertRaises(
            storeapi.errors.KeyAlreadyRegisteredError, self.run_command, ["create-key"]
        )

        self.assertThat(
            str(raised), Equals("You have already registered a key named 'default'")
        )

    def test_create_key_already_registered(self):
        self.fake_store_account_info.mock.side_effect = [
            {
                "account_id": "abcd",
                "account_keys": [
                    {
                        "name": "new-key",
                        "public-key-sha3-384": (get_sample_key("default")["sha3-384"]),
                    }
                ],
            }
        ]
        raised = self.assertRaises(
            storeapi.errors.KeyAlreadyRegisteredError,
            self.run_command,
            ["create-key", "new-key"],
        )

        self.assertThat(
            str(raised), Equals("You have already registered a key named 'new-key'")
        )
        self.fake_check_call.mock.assert_not_called()

    def test_create_key(self):
        self.fake_store_account_info.mock.side_effect = [
            {
                "account_id": "abcd",
                "account_keys": [
                    {
                        "name": "old-key",
                        "public-key-sha3-384": (get_sample_key("default")["sha3-384"]),
                    }
                ],
            }
        ]

        result = self.run_command(["create-key", "new-key"])

        self.assertThat(result.exit_code, Equals(0))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import Contains, DirExists, FileExists, Not

from tests import integration


class CleanPullStepPulledTestCase(integration.TestCase):
    def setUp(self):
        super().setUp()

        self.copy_project_to_cwd("independent-parts")
        self.run_snapcraft("pull")
        self.part1_sourcedir = os.path.join(self.parts_dir, "part1", "src")
        self.part2_sourcedir = os.path.join(self.parts_dir, "part2", "src")

    def assert_files_exist(self):
        self.assertThat(os.path.join(self.part1_sourcedir, "file1"), FileExists())
        self.assertThat(os.path.join(self.part2_sourcedir, "file2"), FileExists())

    def test_clean_pull_step(self):
        self.assert_files_exist()

        self.run_snapcraft(["clean", "--step=pull"])
        self.assertThat(self.part1_sourcedir, Not(DirExists()))
        self.assertThat(self.part2_sourcedir, Not(DirExists()))

        # Now try to pull again
        self.run_snapcraft("pull")
        self.assert_files_exist()

    def test_clean_pull_step_single_part(self):
        self.assert_files_exist()

        self.run_snapcraft(["clean", "part1", "--step=pull"])
        self.assertThat(self.part1_sourcedir, Not(DirExists()))
        self.assertThat(os.path.join(self.part2_sourcedir, "file2"), FileExists())

        # Now try to pull again
        self.run_snapcraft("pull")
        self.assert_files_exist()


class CleanPullStepPrimedTestCase(integration.TestCase):
    def setUp(self):
        super().setUp()

        self.copy_project_to_cwd("independent-parts")
        self.run_snapcraft("prime")

        self.snap_bindir = os.path.join(self.prime_dir, "bin")
        self.stage_bindir = os.path.join(self.stage_dir, "bin")
        self.parts = {}
        for part in ["part1", "part2"]:
            partdir = os.path.join(self.parts_dir, part)
            self.parts[part] = {
                "partdir": partdir,
                "sourcedir": os.path.join(partdir, "src"),
                "builddir": os.path.join(partdir, "build"),
                "installdir": os.path.join(partdir, "install"),
                "bindir": os.path.join(partdir, "install", "bin"),
            }

    def assert_files_exist(self):
        for d in ["builddir", "bindir", "sourcedir"]:
            self.assertThat(os.path.join(self.parts["part1"][d], "file1"), FileExists())
            self.assertThat(os.path.join(self.parts["part2"][d], "file2"), FileExists())

        self.assertThat(os.path.join(self.snap_bindir, "file1"), FileExists())
        self.assertThat(os.path.join(self.snap_bindir, "file2"), FileExists())
        self.assertThat(os.path.join(self.stage_bindir, "file1"), FileExists())
        self.assertThat(os.path.join(self.stage_bindir, "file2"), FileExists())

    def test_clean_pull_step(self):
        self.assert_files_exist()

        output = self.run_snapcraft(["clean", "--step=pull"], debug=False)
        self.assertThat(self.stage_dir, Not(DirExists()))
        self.assertThat(self.prime_dir, Not(DirExists()))

        for part_name, part in self.parts.items():
            self.assertThat(part["builddir"], Not(DirExists()))
            self.assertThat(part["installdir"], Not(DirExists()))
            self.assertThat(part["sourcedir"], Not(DirExists()))

        # Assert that the priming and staging areas were removed wholesale, not
        # a part at a time (since we didn't specify any parts).
        output = output.strip()
        self.expectThat(
            output,
            Contains(
                "Cleaning up priming area\n"
                "Cleaning up staging area\n"
                "Cleaning up parts directory"
            ),
        )

        # Now try to prime again
        self.run_snapcraft("prime")
        self.assert_files_exist()

    def test_clean_pull_step_single_part(self):
        self.assert_files_exist()

        self.run_snapcraft(["clean", "part1", "--step=pull"])
        self.assertThat(os.path.join(self.stage_bindir, "file1"), Not(FileExists()))
        self.assertThat(os.path.join(self.stage_bindir, "file2"), FileExists())
        self.assertThat(os.path.join(self.snap_bindir, "file1"), Not(FileExists()))
        self.assertThat(os.path.join(self.snap_bindir, "file2"), FileExists())

        self.assertThat(self.parts["part1"]["builddir"], Not(DirExists()))
        self.assertThat(self.parts["part1"]["installdir"], Not(DirExists()))
        self.assertThat(self.parts["part1"]["sourcedir"], Not(DirExists()))

        self.assertThat(
            os.path.join(self.parts["part2"]["builddir"], "file2"), FileExists()
        )
        self.assertThat(
            os.path.join(self.parts["part2"]["bindir"], "file2"), FileExists()
        )
        self.assertThat(
            os.path.join(self.parts["part2"]["sourcedir"], "file2"), FileExists()
        )

        # Now try to prime again
        self.run_snapcraft("prime")
        self.assert_files_exist()
